//--------------------------------------------------------------
// File     : stm32_ub_spi4.c
// Datum    : 20.12.2014
// Version  : 1.1b (STM32-Basic-Version)
// Autor    : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
// CPU      : STM32F429
// IDE      : CooCox CoIDE 1.7.4
// GCC      : 4.7 2012q4
// Module   : GPIO, SPI
// Funktion : SPI-LoLevel-Funktionen (SPI-4) Full-Duplex-Mode
//
// Hinweis  : mgliche Pinbelegungen
//            SPI4 : SCK :[PE2, PE12] 
//                   MOSI:[PE6, PE14]
//                   MISO:[PE5, PE13]
//--------------------------------------------------------------

//--------------------------------------------------------------
// Includes
//--------------------------------------------------------------
#include "stm32_ub_spi4.h"


//--------------------------------------------------------------
// Definition von SPI4
//--------------------------------------------------------------
SPI4_DEV_t SPI4DEV = {
// PORT , PIN      , Clock              , Source 
  {GPIOE,GPIO_Pin_2,RCC_AHB1Periph_GPIOE,GPIO_PinSource2}, // SCK an PE2
  {GPIOE,GPIO_Pin_6,RCC_AHB1Periph_GPIOE,GPIO_PinSource6}, // MOSI an PE6
  {GPIOE,GPIO_Pin_5,RCC_AHB1Periph_GPIOE,GPIO_PinSource5}, // MISO an PE5
};



//--------------------------------------------------------------
// Init von SPI4 (im Full-Duplex-Mode) 
// Return_wert :
//  -> ERROR   , wenn SPI schon mit anderem Mode initialisiert
//  -> SUCCESS , wenn SPI init ok war
//--------------------------------------------------------------
ErrorStatus UB_SPI4_Init(SPI4_Mode_t mode, uint16_t vorteiler)
{
  ErrorStatus ret_wert=ERROR;
  GPIO_InitTypeDef  GPIO_InitStructure;
  SPI_InitTypeDef  SPI_InitStructure;

  // SPI-Clock enable
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_SPI4, ENABLE);

  // Clock Enable der Pins
  RCC_AHB1PeriphClockCmd(SPI4DEV.SCK.CLK, ENABLE);
  RCC_AHB1PeriphClockCmd(SPI4DEV.MOSI.CLK, ENABLE);
  RCC_AHB1PeriphClockCmd(SPI4DEV.MISO.CLK, ENABLE);

  // SPI Alternative-Funktions mit den IO-Pins verbinden
  GPIO_PinAFConfig(SPI4DEV.SCK.PORT, SPI4DEV.SCK.SOURCE, GPIO_AF_SPI4);
  GPIO_PinAFConfig(SPI4DEV.MOSI.PORT, SPI4DEV.MOSI.SOURCE, GPIO_AF_SPI4);
  GPIO_PinAFConfig(SPI4DEV.MISO.PORT, SPI4DEV.MISO.SOURCE, GPIO_AF_SPI4);

  // SPI als Alternative-Funktion mit PullDown
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd  = GPIO_PuPd_NOPULL;

  // SCK-Pin
  GPIO_InitStructure.GPIO_Pin = SPI4DEV.SCK.PIN;
  GPIO_Init(SPI4DEV.SCK.PORT, &GPIO_InitStructure);
  // MOSI-Pin
  GPIO_InitStructure.GPIO_Pin = SPI4DEV.MOSI.PIN;
  GPIO_Init(SPI4DEV.MOSI.PORT, &GPIO_InitStructure);
  // MISO-Pin
  GPIO_InitStructure.GPIO_Pin = SPI4DEV.MISO.PIN;
  GPIO_Init(SPI4DEV.MISO.PORT, &GPIO_InitStructure);

  SPI_I2S_DeInit(SPI4);

  // SPI-Konfiguration
  SPI_InitStructure.SPI_Direction = SPI_Direction_2Lines_FullDuplex;
  SPI_InitStructure.SPI_Mode = SPI_Mode_Master;
  SPI_InitStructure.SPI_DataSize = SPI_DataSize_8b;
  if((mode==SPI4_MODE_0_MSB) || (mode==SPI4_MODE_0_LSB)) {
    SPI_InitStructure.SPI_CPOL = SPI_CPOL_Low;
    SPI_InitStructure.SPI_CPHA = SPI_CPHA_1Edge;
  }else if((mode==SPI4_MODE_1_MSB) || (mode==SPI4_MODE_1_LSB)) {
    SPI_InitStructure.SPI_CPOL = SPI_CPOL_Low;
    SPI_InitStructure.SPI_CPHA = SPI_CPHA_2Edge;
  }else if((mode==SPI4_MODE_2_MSB) || (mode==SPI4_MODE_2_LSB)) {
    SPI_InitStructure.SPI_CPOL = SPI_CPOL_High;
    SPI_InitStructure.SPI_CPHA = SPI_CPHA_1Edge;
  }else {
    SPI_InitStructure.SPI_CPOL = SPI_CPOL_High;
    SPI_InitStructure.SPI_CPHA = SPI_CPHA_2Edge;
  } 
  SPI_InitStructure.SPI_NSS = SPI_NSS_Soft;
  SPI_InitStructure.SPI_BaudRatePrescaler = vorteiler;

  if((mode==SPI4_MODE_0_MSB) || (mode==SPI4_MODE_1_MSB) ||
     (mode==SPI4_MODE_2_MSB) || (mode==SPI4_MODE_3_MSB)) {
    SPI_InitStructure.SPI_FirstBit = SPI_FirstBit_MSB;
  }
  else {
    SPI_InitStructure.SPI_FirstBit = SPI_FirstBit_LSB;
  }

  SPI_InitStructure.SPI_CRCPolynomial = 7;
  SPI_Init(SPI4, &SPI_InitStructure); 

  // SPI enable
  SPI_Cmd(SPI4, ENABLE); 

  ret_wert=SUCCESS;

  return(ret_wert);
}


//--------------------------------------------------------------
// sendet und empfngt ein Byte per SPI4 (im Full-Duplex-Mode)
// ChipSelect-Signal muss von rufender Funktion gemacht werden
//--------------------------------------------------------------
uint8_t UB_SPI4_SendByte(uint8_t wert)
{ 
  uint8_t ret_wert=0;

  // warte bis senden fertig
  while (SPI_I2S_GetFlagStatus(SPI4, SPI_I2S_FLAG_TXE) == RESET);

  // byte senden
  SPI_I2S_SendData(SPI4, wert);

  // warte bis empfang fertig
  while (SPI_I2S_GetFlagStatus(SPI4, SPI_I2S_FLAG_RXNE) == RESET); 

  // Daten einlesen
  ret_wert=SPI_I2S_ReceiveData(SPI4);

  return(ret_wert);
}
