/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


//------------------------------------------------------------
// Changelog
//------------------------------------------------------------
//  [V:0.1 / 04.09.2013] :
// Author   : Adam Dunkels
// Web      : http://dunkels.com/adam/ubasic/
//
// original Version from Adam Dunkels without changings
//------------------------------------------------------------
//  [V:2.0 / 20.12.2014] :
// Author   : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
// Port for STM32F429
//
// modified : removed all debug statements
// modified : replaced exit() statement with variable "failed"
// added    : function to read status of variable "failed"
// added    : function to read line-nr with failure
// bugfix   : variable "a" is permitted
// bugfix   : if/then/else support with :
//             "print","let","goto","gosub","for"
//             e.g.  "10 if i>3 then k=10 else k=20"
// modified : Token "then" not necessary
//             e.g.  "10 if i>3 k=10 else k=20"
// modified : negativ values possible
//             e.g   "10 a=-4"
//                   "20 b=-(c+3)"
// added    : hex values supported
//             e.g.  10 a=0xf4
// added    : XOR function :
//             e.g.  "10 a=8^4"
// modified : removed unused Token "CALL"
// addded   : new Token "LINENUM" to find line numbers
// modified : variable type from char to integer
// modified : faster "jump_linenum" function
// added    : decrement support in "for/next"
//             e.g   "10 for i=10 to 2"
//                   "20 next i"
// added    : new Token "step" (if needed)
//             e.g   "10 for i=10 to 20 step 3"
//                   "20 next i"
// added    : ignore newline in PRINT with ";"
// added    : insert tabulator in PRINT with ","
// added    : string support
//             e.g   "10 a$="Hello"
// added    : int array support
//             e.g   "10 DIM A(10)"
//                   "20 A(3)=17"
// added    : str array support
//             e.g.  "10 DIM A$(10)"
//                   "20 A$(3)="Hallo"
// added    : data array support
//             e.g.  "10 DIM A#(10)"
//                   "20 A#(3)=17"
// added    : 3stage pre-parse function (see tokenizer)
// added    : many new basic commands for STM32F429
//------------------------------------------------------------



#include "ubasic.h"
#include "tokenizer.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


//------------------------------------------------------------
// sine table (-10000 to +10000)
//------------------------------------------------------------
static const int SineTable[360] = {
  0, 175, 349, 523, 698, 872, 1045, 1219, 1392,
  1564, 1736, 1908, 2079, 2250, 2419, 2588, 2756, 2924,
  3090, 3256, 3420, 3584, 3746, 3907, 4067, 4226, 4384,
  4540, 4695, 4848, 5000, 5150, 5299, 5446, 5592, 5736,
  5878, 6018, 6157, 6293, 6428, 6561, 6691, 6820, 6947,
  7071, 7193, 7314, 7431, 7547, 7660, 7771, 7880, 7986,
  8090, 8192, 8290, 8387, 8480, 8572, 8660, 8746, 8829,
  8910, 8988, 9063, 9135, 9205, 9272, 9336, 9397, 9455,
  9511, 9563, 9613, 9659, 9703, 9744, 9781, 9816, 9848,
  9877, 9903, 9925, 9945, 9962, 9976, 9986, 9994, 9998,
  10000, 9998, 9994, 9986, 9976, 9962, 9945, 9925, 9903,
  9877, 9848, 9816, 9781, 9744, 9703, 9659, 9613, 9563,
  9511, 9455, 9397, 9336, 9272, 9205, 9135, 9063, 8988,
  8910, 8829, 8746, 8660, 8572, 8480, 8387, 8290, 8192,
  8090, 7986, 7880, 7771, 7660, 7547, 7431, 7314, 7193,
  7071, 6947, 6820, 6691, 6561, 6428, 6293, 6157, 6018,
  5878, 5736, 5592, 5446, 5299, 5150, 5000, 4848, 4695,
  4540, 4384, 4226, 4067, 3907, 3746, 3584, 3420, 3256,
  3090, 2924, 2756, 2588, 2419, 2250, 2079, 1908, 1736,
  1564, 1392, 1219, 1045, 872, 698, 523, 349, 175,
  0, -175, -349, -523, -698, -872, -1045, -1219, -1392,
  -1564, -1736, -1908, -2079, -2250, -2419, -2588, -2756, -2924,
  -3090, -3256, -3420, -3584, -3746, -3907, -4067, -4226, -4384,
  -4540, -4695, -4848, -5000, -5150, -5299, -5446, -5592, -5736,
  -5878, -6018, -6157, -6293, -6428, -6561, -6691, -6820, -6947,
  -7071, -7193, -7314, -7431, -7547, -7660, -7771, -7880, -7986,
  -8090, -8192, -8290, -8387, -8480, -8572, -8660, -8746, -8829,
  -8910, -8988, -9063, -9135, -9205, -9272, -9336, -9397, -9455,
  -9511, -9563, -9613, -9659, -9703, -9744, -9781, -9816, -9848,
  -9877, -9903, -9925, -9945, -9962, -9976, -9986, -9994, -9998,
  -10000, -9998, -9994, -9986, -9976, -9962, -9945, -9925, -9903,
  -9877, -9848, -9816, -9781, -9744, -9703, -9659, -9613, -9563,
  -9511, -9455, -9397, -9336, -9272, -9205, -9135, -9063, -8988,
  -8910, -8829, -8746, -8660, -8572, -8480, -8387, -8290, -8192,
  -8090, -7986, -7880, -7771, -7660, -7547, -7431, -7314, -7193,
  -7071, -6947, -6820, -6691, -6561, -6428, -6293, -6157, -6018,
  -5878, -5736, -5592, -5446, -5299, -5150, -5000, -4848, -4695,
  -4540, -4384, -4226, -4067, -3907, -3746, -3584, -3420, -3256,
  -3090, -2924, -2756, -2588, -2419, -2250, -2079, -1908, -1736,
  -1564, -1392, -1219, -1045, -872, -698, -523, -349, -175
};


//------------------------------------------------------------
// pointer to BASIC-RAM
//------------------------------------------------------------
char *program_ptr;
char *memory_ptr;


//------------------------------------------------------------
// buffer for single string
//------------------------------------------------------------
#define MAX_STRINGLEN 85
static char string[MAX_STRINGLEN];


//------------------------------------------------------------
// GOSUB Buffer
//------------------------------------------------------------
#define MAX_GOSUB_STACK_DEPTH 10
static int gosub_stack[MAX_GOSUB_STACK_DEPTH];
static int gosub_stack_ptr;


//------------------------------------------------------------
// FOR Buffer
//------------------------------------------------------------
struct for_state {
  int line_after_for;
  int for_variable;
  int to;
  int step;
};
#define MAX_FOR_STACK_DEPTH 4
static struct for_state for_stack[MAX_FOR_STACK_DEPTH];
static int for_stack_ptr;

//------------------------------------------------------------
// int-variables
//------------------------------------------------------------
#define MAX_VARNUM 26  // dont change this value
static int int_var[MAX_VARNUM];

//------------------------------------------------------------
// added by UB
// string-variables
//------------------------------------------------------------
static char str_var[MAX_VARNUM][MAX_STRINGLEN];
char strbuf1[MAX_STRINGLEN]; // str buffer
char strbuf2[MAX_STRINGLEN]; // str buffer


//------------------------------------------------------------
// added by UB
// int-variables-arrays (in external SDRAM)
// bytes = size * 4
//------------------------------------------------------------
#define MAX_INT_ARRAY_SIZE 8*1024 // 32kByte
struct int_var_array_t {
  int lenght;
  int start_adr;
};
static struct int_var_array_t int_var_array[MAX_VARNUM];
int int_var_array_data[MAX_INT_ARRAY_SIZE] __attribute__((section(".HeapMemSection")));
static int int_var_array_ptr;


//------------------------------------------------------------
// added by UB
// string-variables-arrays (in external SDRAM)
// bytes = size * 85
//------------------------------------------------------------
#define MAX_STR_ARRAY_SIZE 400 // 33kByte
struct str_var_array_t {
  int lenght;
  int start_adr;
};
static struct str_var_array_t str_var_array[MAX_VARNUM];
char str_var_array_data[MAX_STR_ARRAY_SIZE][MAX_STRINGLEN] __attribute__((section(".HeapMemSection")));
static int str_var_array_ptr;


//------------------------------------------------------------
// added by UB
// char-variables-arrays (in external SDRAM)
// bytes = size * 1
//------------------------------------------------------------
#define MAX_DATA_ARRAY_SIZE 64*1024 // 64kByte
struct data_var_array_t {
  int lenght;
  int start_adr;
};
static struct data_var_array_t data_var_array[MAX_VARNUM];
char data_var_array_data[MAX_DATA_ARRAY_SIZE] __attribute__((section(".HeapMemSection")));
static int data_var_array_ptr;
static int akt_data_array_var;
static int akt_data_array_pos;


//------------------------------------------------------------
static int ended;
static int failed, akt_linenum;       // added by UB
static int akt_cmd;                   // added by UB
static int akt_xp, akt_yp;            // added by UB
static int v24_xp;                    // added by UB


//------------------------------------------------------------
static int expr(void);
static char *strexpr(void); // added by UB
static void linenum_statement(void);
static void statement(void);
static void skipelse(void);  // added by UB
static void string_to_cmd(char *ptr, char mode); // added by UB


//------------------------------------------------------------
// added by UB
// basic commands with integer return value
//------------------------------------------------------------
static int cmd_abs(void);


//------------------------------------------------------------
// added by UB
// basic commands with string return value
//------------------------------------------------------------
static char *cmd_rightstr(void);


//------------------------------------------------------------
// init
// mode : 0=reset variable
//        1=variable dont change (for inline interpreter)
//------------------------------------------------------------
void ubasic_init(char *program, int mode)
{
  int n;

  program_ptr = program;
  memory_ptr = program;
  for_stack_ptr = gosub_stack_ptr = 0;
  ended = 0;
  failed =0;
  akt_linenum =0;
  akt_cmd=0;
  akt_xp=0;
  akt_yp=0;
  v24_xp=0;
  // randomize
  n=UB_Systick_BasicReadTIC();
  srand(n);
  rand();
  // pre-parse the basic-program
  n=tokenizer_pre_parse(program);
  if(n!=0) {
    failed=UBASIC_PREPARSE_ERR;
    akt_linenum=tokenizer_linenum();
  }

  INTERPRETER.init_port[0]=1; // init already ok
  INTERPRETER.init_port[1]=0;
  INTERPRETER.init_port[2]=1; // init already ok
  INTERPRETER.init_port[3]=1; // init already ok
  INTERPRETER.init_port[4]=0;
  INTERPRETER.init_adc=0;

  if(mode==0) {
    for(n=0;n<MAX_VARNUM;n++) {
      // integer variables
      int_var[n]=0;
      // string variables
      str_var[n][0]=0x00;
      // integer arrays
      int_var_array[n].lenght=0;
      int_var_array[n].start_adr=0;
      // string arrays
      str_var_array[n].lenght=0;
      str_var_array[n].start_adr=0;
      // data arrays
      data_var_array[n].lenght=0;
      data_var_array[n].start_adr=0;
    }
    int_var_array_ptr=0;
    str_var_array_ptr=0;
    data_var_array_ptr=0;
    akt_data_array_var=-1;
    akt_data_array_pos=0;

    // integer data
    for(n=0;n<MAX_INT_ARRAY_SIZE;n++) {
      int_var_array_data[n]=0;
    }
    // string data
    for(n=0;n<MAX_STR_ARRAY_SIZE;n++) {
      str_var_array_data[n][0]=0x00;
    }
    // char data
    for(n=0;n<MAX_DATA_ARRAY_SIZE;n++) {
      data_var_array_data[n]=0;
    }
  }
}
//------------------------------------------------------------
// check if token on the right place
//------------------------------------------------------------
static void accept(int token)
{
  if(token != tokenizer_token()) {
    tokenizer_error_print();
    failed=UBASIC_UNEXPECTED_TOKEN;
  }
  else {
    tokenizer_next();
  }
}
//------------------------------------------------------------
// return a value from a integer variable 'a' to 'z'
//------------------------------------------------------------
static int varfactor(void)
{
  int r;
  r = ubasic_get_int_var(tokenizer_variable_num());
  accept(TOKENIZER_VARIABLE);
  return r;
}
//------------------------------------------------------------
// return a value from a integer var_array 'A(n)' to 'Z(n)'
//------------------------------------------------------------
static int varfactor_array(int varnum, int varpos)
{
  int r;
  r = ubasic_get_int_var_array(varnum ,varpos);
  return r;
}
//------------------------------------------------------------
// return a value from a char data_array 'A#(n)' to 'Z#(n)'
//------------------------------------------------------------
static int varfactor_dataarray(int varnum, int varpos)
{
  int r;
  r = ubasic_get_data_var_array(varnum ,varpos);
  return r;
}
//------------------------------------------------------------
// return factor
// insert all cases in "void tokenizer_set_expression(void)"
//------------------------------------------------------------
static int factor(void)
{
  int r=0,p=0,q=0;
  int mode=0;

  if(tokenizer_token()==TOKENIZER_MINUS) {
    // negativ
    mode=1;
    accept(TOKENIZER_MINUS);
  }

  switch(tokenizer_token()) {
    case TOKENIZER_NUMBER:  // '1234'
      r = tokenizer_num();
      accept(TOKENIZER_NUMBER);
    break;
    case TOKENIZER_HEXNUM:  // '0x1234'
      r = tokenizer_hexnum();
      accept(TOKENIZER_HEXNUM);
    break;
    case TOKENIZER_LEFTPAREN: // '(r)'
      accept(TOKENIZER_LEFTPAREN);
      r = expr();
      accept(TOKENIZER_RIGHTPAREN);
    break;
    case TOKENIZER_VARIABLE: // 'a'
      r = varfactor();
    break;
    case TOKENIZER_VARARRAY:  // 'A(q)'
      p=tokenizer_variable_numBig();
      accept(TOKENIZER_VARARRAY);
      accept(TOKENIZER_LEFTPAREN);
      q=expr();
      accept(TOKENIZER_RIGHTPAREN);
      r = varfactor_array(p,q);
    break;
    case TOKENIZER_DATAARRAY:  // 'A#(q)'
      p=tokenizer_variable_numBig();
      accept(TOKENIZER_DATAARRAY);
      accept(TOKENIZER_LEFTPAREN);
      q=expr();
      accept(TOKENIZER_RIGHTPAREN);
      r = varfactor_dataarray(p,q);
    break;
    case TOKENIZER_ABS : // 'ABS'
      r=cmd_abs();
    break;
    default:
      failed=UBASIC_UNEXPECTED_TOKEN;
    break;
  }

  if(mode==1) r=0-r; // negation

  return r;
}
//------------------------------------------------------------
// calculate
// multiplication '*'
// division       '/'
// modulo         '%'
//------------------------------------------------------------
static int term(void)
{
  int f1, f2;
  int op;

  f1 = factor();
  op = tokenizer_token();
  while(op == TOKENIZER_ASTR ||
	op == TOKENIZER_SLASH ||
	op == TOKENIZER_MOD) {
    tokenizer_next();
    f2 = factor();
    switch(op) {
    case TOKENIZER_ASTR: // '*'
      f1 = f1 * f2;
      break;
    case TOKENIZER_SLASH:  // '/'
      f1 = f1 / f2;
      break;
    case TOKENIZER_MOD:  // '%'
      f1 = f1 % f2;
      break;
    }
    op = tokenizer_token();
  }
  return f1;
}
//------------------------------------------------------------
// calculate
// plus        '+'
// minus       '-'
// logical AND '&'
// logical OR  '|'
// logical XOR '^'
//------------------------------------------------------------
static int expr(void)
{
  int t1, t2;
  int op;

  t1 = term();
  op = tokenizer_token();
  while(op == TOKENIZER_PLUS ||
    op == TOKENIZER_MINUS ||
    op == TOKENIZER_AND ||
    op == TOKENIZER_OR ||
    op == TOKENIZER_XOR) {
    tokenizer_next();
    t2 = term();
    switch(op) {
    case TOKENIZER_PLUS:  // '+'
      t1 = t1 + t2;
      break;
    case TOKENIZER_MINUS:  // '-'
      t1 = t1 - t2;
      break;
    case TOKENIZER_AND:  // '&'
      t1 = t1 & t2;
      break;
    case TOKENIZER_OR:  // '|'
      t1 = t1 | t2;
      break;
    case TOKENIZER_XOR:  // '^'
      t1 = t1 ^ t2;
      break;
    }
    op = tokenizer_token();
  }
  return t1;
}
//------------------------------------------------------------
// read one string
// insert all cases in "void tokenizer_set_expression(void)"
//------------------------------------------------------------
static char *strpart(void)
{
  int p,q;
  char *r;

  switch(tokenizer_token()) {
    case TOKENIZER_STRING : // 'Test'
      tokenizer_string(string, sizeof(string));
      r=string;
      tokenizer_next();
    break;
    case TOKENIZER_STRINGVAR : // 'a$'
      ubasic_get_str_var(tokenizer_variable_num());
      r=string;
      tokenizer_next();
    break;
    case TOKENIZER_STRINGARRAY : // 'A$(p)'
      p=tokenizer_variable_numBig();
      accept(TOKENIZER_STRINGARRAY);
      accept(TOKENIZER_LEFTPAREN);
      q=expr();
      accept(TOKENIZER_RIGHTPAREN);
      ubasic_get_str_var_array(p,q);
      r=string;
    break;
    case TOKENIZER_DATAARRAY : // 'A#(p)'
      p=tokenizer_variable_numBig();
      accept(TOKENIZER_DATAARRAY);
      accept(TOKENIZER_LEFTPAREN);
      q=expr();
      accept(TOKENIZER_RIGHTPAREN);
      ubasic_get_datas_var_array(p,q);
      r=string;
    break;
    case TOKENIZER_RIGHTSTR : // 'RIGHT$'
      r=cmd_rightstr();
    break;
    default:
      failed=UBASIC_UNEXPECTED_TOKEN;
  }

  return r;
}
//------------------------------------------------------------
// cat strings
// plus        '+'
//------------------------------------------------------------
static char *strexpr(void)
{
  char s1[MAX_STRINGLEN+1];
  char s2[MAX_STRINGLEN+1];
  int len1,len2;
  int op;

  strcpy(s1, strpart());
  len1=strlen(s1);
  op = tokenizer_token();
  while(op == TOKENIZER_PLUS) { // '+'
    tokenizer_next();
    strcpy(s2, strpart());
    len2=strlen(s2);
    if(len1+len2<MAX_STRINGLEN) {
      strcat(s1, s2);
      len1=strlen(s1);
    }
    else failed=UBASIC_OVERFLOW;

    op = tokenizer_token();
  }
  strcpy(strbuf1, s1);
  return strbuf1;
}
//------------------------------------------------------------
// check relation
// equal  '='
// smaler '<'
// bigger '>'
//------------------------------------------------------------
static int relation(void)
{
  int r1, r2;
  int op;

  r1 = expr();
  op = tokenizer_token();
  while(op == TOKENIZER_LT ||
	op == TOKENIZER_GT ||
	op == TOKENIZER_EQ) {
    tokenizer_next();
    r2 = expr();
    switch(op) {
    case TOKENIZER_LT: // '<'
      r1 = r1 < r2;
      break;
    case TOKENIZER_GT:  // '>'
      r1 = r1 > r2;
      break;
    case TOKENIZER_EQ:  // '='
      r1 = r1 == r2;
      break;
    }
    op = tokenizer_token();
  }
  return r1;
}
//------------------------------------------------------------
// check string relation
// equal  '='  "A"="A"
// smaler '<'  "A"<"B"
// bigger '>'  "B">"A"
//------------------------------------------------------------
static int strrelation(void)
{
  char s1[MAX_STRINGLEN+1];
  char s2[MAX_STRINGLEN+1];
  int r1=0, r2;
  int op;

  strcpy(s1, strexpr());
  op = tokenizer_token();
  while(op == TOKENIZER_LT ||
	op == TOKENIZER_GT ||
	op == TOKENIZER_EQ) {
    tokenizer_next();
    strcpy(s2, strexpr());
    r2=strcmp(s1, s2);
    switch(op) {
    case TOKENIZER_LT:  // '<'
      if(r2<0) r1=1;
      break;
    case TOKENIZER_GT:  // '>'
      if(r2>0) r1=1;
      break;
    case TOKENIZER_EQ:  // '='
      if(r2==0) r1=1;
      break;
    }
    op = tokenizer_token();
  }

  return r1;
}
//------------------------------------------------------------
// jump to line-nummer (linenum)
//------------------------------------------------------------
static void jump_linenum(int linenum)
{
  int ok=0,akt_token,aktline;

  // try to jump from a pre-parsed linenumber
  ok=tokenizer_jumpfast(linenum);
  if(ok==1) return;

  // search the linenumber manually
  tokenizer_init(program_ptr);

  ok=0;
  do {
    akt_token=tokenizer_token();
    if(akt_token==TOKENIZER_ERROR) {
      ok=2;
    }
    else if(akt_token==TOKENIZER_ENDOFINPUT) {
      ok=1;
    }
    else if(akt_token==TOKENIZER_LINENUM) {
      aktline=tokenizer_num();
      if(aktline==linenum)  {
        ok=3;
      }
      else {
        tokenizer_skipline();
        tokenizer_next();
      }
    }
    else {
      ok=4;
    }
  }while(ok==0);
}
//------------------------------------------------------------
// Statement : "GOTO"
// Syntax    : GOTO<uint>
// Return    : <void>
//------------------------------------------------------------
static void goto_statement(void)
{
  accept(TOKENIZER_GOTO);
  jump_linenum(tokenizer_num());
}
//------------------------------------------------------------
// Statement : "PRINT"
// Syntax    : PRINT<printable>[<tr><printable>...]
// Return    : <void>
//------------------------------------------------------------
static void print_statement(void)
{
  int newline=1;

  accept(TOKENIZER_PRINT);
  do {
    if(tokenizer_token() == TOKENIZER_STRING) {
      tokenizer_string(string, sizeof(string));
      string_to_cmd(string,0);
      newline=1;
      tokenizer_next();
    } else if(tokenizer_is_expr()==1) { // integer expression
      sprintf(string,"%d",expr());
      string_to_cmd(string,0);
      newline=1;
    } else if(tokenizer_is_expr()==2) { // string expression
      sprintf(string,"%s",strexpr());
      string_to_cmd(string,0);
      newline=1;
    } else if(tokenizer_is_expr()==3) { // data-array as integer expression
      sprintf(string,"%d",expr());
      string_to_cmd(string,0);
      newline=1;
    } else if(tokenizer_token()==TOKENIZER_LEFTPAREN) {
      sprintf(string,"%d",expr());
      string_to_cmd(string,0);
      newline=1;
    } else if(tokenizer_token() == TOKENIZER_COMMA) {
      string_to_cmd("",1);
      newline=0;
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_SEMICOLON) {
      newline=0;
      tokenizer_next();
    } else {
      break;
    }
  } while(tokenizer_token() != TOKENIZER_CR &&
	  tokenizer_token() != TOKENIZER_ENDOFINPUT);

  if(newline==1) string_to_cmd("",2);

  if(tokenizer_token()!=TOKENIZER_ELSE) {
    tokenizer_next();
  }
}
//------------------------------------------------------------
// this is a subfunction from "PRINT"
// mode : 0 = normal
// mode : 1 = TAB
// mode : 2 = Newline
//------------------------------------------------------------
static void string_to_cmd(char *ptr, char mode)
{
  int str_len,max;

  // stringlen
  str_len=strlen(ptr);

  if(akt_cmd==0) {
    // print fuer LCD
    #if PLATFORM==1
    // print fuer PC
    if(mode==0) {
      printf("%s", ptr);
      v24_xp+=(str_len*8);
    }
    else if(mode==1) {
      max=TAB_POS-(v24_xp%TAB_POS);
      for(str_len=0;str_len<max;str_len+=8) {
        printf(" ");
      }
      v24_xp+=max;
    }
    else if(mode==2) {
      printf("\n");
      v24_xp=0;
    }
    #endif
    #if PLATFORM==2
    if(mode==0) {
      UB_Font_BasicString(akt_xp,akt_yp,ptr,UBASIC.font_ptr,UBASIC.vg,UBASIC.bg);
      akt_xp+=(str_len*UBASIC.font_ptr->width);
      if(akt_xp>LCD_MAXY) akt_xp=LCD_MAXY;
    }
    else if(mode==1) {
      max=TAB_POS-(akt_xp%TAB_POS);
      akt_xp+=max;
      if(akt_xp>LCD_MAXY) akt_xp=LCD_MAXY;
    }
    else if(mode==2) {
      akt_xp=0;
      akt_yp+=UBASIC.font_ptr->height;
      if(akt_yp>LCD_MAXX) akt_yp=LCD_MAXX;
    }
    #endif
  }
  else if(akt_cmd==1) {
    #if PLATFORM==2
    // print fuer UART
    if(mode==0) {
      UB_Uart_SendString(COM1,ptr,NONE);
      v24_xp+=(str_len*8);
    }
    else if(mode==1) {
      max=TAB_POS-(v24_xp%TAB_POS);
      for(str_len=0;str_len<max;str_len+=8) {
        UB_Uart_SendString(COM1," ",NONE);
      }
      v24_xp+=max;
    }
    else if(mode==2){
      UB_Uart_SendString(COM1,ptr,CRLF);
      v24_xp=0;
    }
    #endif
  }
}
//------------------------------------------------------------
// Statement : "IF"
// Syntax    : IF<int_expr><rel><int_expr>[THEN]<cmd>[ELSE<cmd>]
//     or    : IF<str_expr><rel><str_expr>[THEN]<cmd>[ELSE<cmd>]
// Return    : <void>
//------------------------------------------------------------
static void if_statement(void)
{
  int r=0;

  accept(TOKENIZER_IF);

  if(tokenizer_is_expr()==1) { // integer expression
    r = relation();
  }
  else if(tokenizer_is_expr()==2) { // string expression
    r = strrelation();
  }
  else if(tokenizer_is_expr()==3) { // data-array as integer expression
    r = relation();
  }
  else if(tokenizer_token()==TOKENIZER_LEFTPAREN) {
    r = relation();
  }
  else failed=UBASIC_UNEXPECTED_TOKEN;

  if(tokenizer_token()==TOKENIZER_THEN) {
    accept(TOKENIZER_THEN);
  }
  if(r) {
    statement();
    if(tokenizer_token()==TOKENIZER_ELSE) {
      skipelse();
    }
  }
  else {
    do {
      tokenizer_next();
    } while(tokenizer_token() != TOKENIZER_ELSE &&
	    tokenizer_token() != TOKENIZER_CR &&
	    tokenizer_token() != TOKENIZER_ENDOFINPUT);

    if(tokenizer_token() == TOKENIZER_ELSE) {
      tokenizer_next();
      statement();
    } else if(tokenizer_token() == TOKENIZER_CR) {
      tokenizer_next();
    }
  }
}
//------------------------------------------------------------
// Statement : "LET"
// Syntax    : [LET]<int_var>=<int_expr>
//    or       [LET]<int_array>=<int_expr>
//    or       [LET]<str_var>=<str_expr>
//    or       [LET]<str_array>=<str_expr>
//    or       [LET]<data_array>=<int_expr>
//    or       [LET]<data_array>=<str_expr>
// Return    : <void>
//------------------------------------------------------------
static void
let_statement(void)
{
  int var,var_index,akt_token;

  akt_token=tokenizer_token();
  if(akt_token==TOKENIZER_VARIABLE) {
    // interger var
    var = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    accept(TOKENIZER_EQ);
    ubasic_set_int_var(var, expr());
  }
  else if(akt_token==TOKENIZER_STRINGVAR) {
    // string var
    var = tokenizer_variable_num();
    accept(TOKENIZER_STRINGVAR);
    accept(TOKENIZER_EQ);
    ubasic_set_str_var(var, strexpr());
  }
  else if(akt_token==TOKENIZER_VARARRAY) {
    // var array
    var = tokenizer_variable_numBig();
    accept(TOKENIZER_VARARRAY);
    accept(TOKENIZER_LEFTPAREN);
    var_index=expr();
    accept(TOKENIZER_RIGHTPAREN);
    accept(TOKENIZER_EQ);
    ubasic_set_int_var_array(var, var_index, expr());
  }
  else if(akt_token==TOKENIZER_STRINGARRAY) {
    // string array
    var = tokenizer_variable_numBig();
    accept(TOKENIZER_STRINGARRAY);
    accept(TOKENIZER_LEFTPAREN);
    var_index=expr();
    accept(TOKENIZER_RIGHTPAREN);
    accept(TOKENIZER_EQ);
    ubasic_set_str_var_array(var, var_index, strexpr());
  }
  else if(akt_token==TOKENIZER_DATAARRAY) {
    // data array
    var = tokenizer_variable_numBig();
    accept(TOKENIZER_DATAARRAY);
    accept(TOKENIZER_LEFTPAREN);
    var_index=expr();
    accept(TOKENIZER_RIGHTPAREN);
    accept(TOKENIZER_EQ);
    if(tokenizer_is_expr()==1) {
      ubasic_set_data_var_array(var, var_index, expr());
    }
    else if(tokenizer_is_expr()==2) {
      ubasic_set_datas_var_array(var, var_index, strexpr());
    }
    else if(tokenizer_is_expr()==3) { // data-array as integer expression
      ubasic_set_data_var_array(var, var_index, expr());
    }
    else failed=UBASIC_UNEXPECTED_TOKEN;
  }
  else failed=UBASIC_UNEXPECTED_TOKEN;

  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
}
//------------------------------------------------------------
// Statement : "GOSUB"
// Syntax    : GOSUB<uint>
// Return    : <void>
//------------------------------------------------------------
static void gosub_statement(void)
{
  int linenum;
  accept(TOKENIZER_GOSUB);
  linenum = tokenizer_num();
  accept(TOKENIZER_NUMBER);

  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }

  if(gosub_stack_ptr < MAX_GOSUB_STACK_DEPTH) {
    gosub_stack[gosub_stack_ptr] = tokenizer_num();
    gosub_stack_ptr++;
    jump_linenum(linenum);
  }
  else failed=UBASIC_OVERFLOW;
}
//------------------------------------------------------------
// Statement : "RETURN"
// Syntax    : RETURN
// Return    : <void>
//------------------------------------------------------------
static void return_statement(void)
{
  accept(TOKENIZER_RETURN);
  if(gosub_stack_ptr > 0) {
    gosub_stack_ptr--;
    jump_linenum(gosub_stack[gosub_stack_ptr]);
  }
  else failed=UBASIC_UNEXPECTED_TOKEN;
}
//------------------------------------------------------------
// Statement : "NEXT"
// Syntax    : NEXT<int_var>
// Return    : <void>
//------------------------------------------------------------
static void next_statement(void)
{
  int var,ist,soll,step;

  accept(TOKENIZER_NEXT);
  var = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  if(for_stack_ptr > 0 &&
    var == for_stack[for_stack_ptr - 1].for_variable) {

    ist=ubasic_get_int_var(var);
    soll=for_stack[for_stack_ptr - 1].to;
    step=for_stack[for_stack_ptr - 1].step;

    if(ist<soll) {
      // count up
      ubasic_set_int_var(var,ist + step);
      if(ubasic_get_int_var(var) <= for_stack[for_stack_ptr - 1].to) {
        jump_linenum(for_stack[for_stack_ptr - 1].line_after_for);
      } else {
        for_stack_ptr--;
        accept(TOKENIZER_CR);
      }
    }
    else if(ist>soll) {
      // count down
      ubasic_set_int_var(var,ist - step);
      if(ubasic_get_int_var(var) >= for_stack[for_stack_ptr - 1].to) {
        jump_linenum(for_stack[for_stack_ptr - 1].line_after_for);
      } else {
        for_stack_ptr--;
        accept(TOKENIZER_CR);
      }
    }
    else {
      for_stack_ptr--;
      accept(TOKENIZER_CR);
    }

  } else {
    accept(TOKENIZER_CR);
  }

}
//------------------------------------------------------------
// Statement : "FOR"
// Syntax    : FOR<int_var>=<int_expr>TO<int_expr>[STEP<int_expr>]
// Return    : <void>
//------------------------------------------------------------
static void for_statement(void)
{
  int for_variable, to, step=1;

  accept(TOKENIZER_FOR);
  for_variable = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  ubasic_set_int_var(for_variable, expr());
  accept(TOKENIZER_TO);
  to = expr();
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    if(tokenizer_token()==TOKENIZER_STEP) {
      accept(TOKENIZER_STEP);
      step = expr();
      if(step<=0) step=1;
      if(tokenizer_token()!=TOKENIZER_ELSE) {
        accept(TOKENIZER_CR);
      }
      else {
        skipelse();
      }
    }
    else {
      accept(TOKENIZER_CR);
    }
  }
  else {
    skipelse();
  }

  if(for_stack_ptr < MAX_FOR_STACK_DEPTH) {
    for_stack[for_stack_ptr].line_after_for = tokenizer_num();
    for_stack[for_stack_ptr].for_variable = for_variable;
    for_stack[for_stack_ptr].to = to;
    for_stack[for_stack_ptr].step = step;

    for_stack_ptr++;
  }
  else failed=UBASIC_OVERFLOW;
}
//------------------------------------------------------------
// Statement : "END"
// Syntax    : END
// Return    : <void>
//------------------------------------------------------------
static void end_statement(void)
{
  accept(TOKENIZER_END);
  ended = 1;
}


//------------------------------------------------------------
// NEW <VOID> STATEMENTS BY UB
//------------------------------------------------------------


//------------------------------------------------------------
// Statement : "REM"
// Syntax    : REM<anything>
// Return    : <void>
//------------------------------------------------------------
static void rem_statement(void)
{
  tokenizer_skipline();
  accept(TOKENIZER_CR);
}
//------------------------------------------------------------
// Statement : "DIM"
// Syntax    : DIM<int_array>(<int_expr>)
//   or      : DIM<str_array>(<int_expr>)
//   or      : DIM<data_array>(<int_expr>)
// Return    : <void>
//------------------------------------------------------------
static void dim_statement(void)
{
  int var=0,value=0,akt_token;

  accept(TOKENIZER_DIM);
  akt_token=tokenizer_token();
  if(akt_token==TOKENIZER_VARARRAY) {
    var = tokenizer_variable_numBig();
    accept(TOKENIZER_VARARRAY);
    accept(TOKENIZER_LEFTPAREN);
    value=expr();
    accept(TOKENIZER_RIGHTPAREN);
    if(failed==0) {
      ubasic_dim_int_array(var, value);
    }
  }
  else if(akt_token==TOKENIZER_STRINGARRAY) {
    var = tokenizer_variable_numBig();
    accept(TOKENIZER_STRINGARRAY);
    accept(TOKENIZER_LEFTPAREN);
    value=expr();
    accept(TOKENIZER_RIGHTPAREN);
    if(failed==0) {
      ubasic_dim_str_array(var, value);
    }
  }
  else if(akt_token==TOKENIZER_DATAARRAY) {
    var = tokenizer_variable_numBig();
    accept(TOKENIZER_DATAARRAY);
    accept(TOKENIZER_LEFTPAREN);
    value=expr();
    accept(TOKENIZER_RIGHTPAREN);
    if(failed==0) {
      ubasic_dim_data_array(var, value);
    }
  }
  else failed=UBASIC_UNEXPECTED_TOKEN;

  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
//------------------------------------------------------------
// Statement : "DATA"
// Syntax    : DATA<int_expr>[,<int_expr>...]
// Return    : <void>
//------------------------------------------------------------
static void data_statement(void)
{
  int value=0, ok=0;

  accept(TOKENIZER_DATA);
  value=expr();
  if((value<0) || (value>255)) failed=UBASIC_INVALID_DATA;
  if(akt_data_array_var<0) failed=UBASIC_OVERFLOW;

  if(failed==0) {
    ubasic_set_data_var_array(akt_data_array_var, akt_data_array_pos, value);
    if(failed==0) akt_data_array_pos++;

    do {
      if(tokenizer_token()==TOKENIZER_COMMA) {
        value=0;
        accept(TOKENIZER_COMMA);
        value=expr();
        if((value<0) || (value>255)) failed=UBASIC_INVALID_DATA;
        if(failed==0) {
          ubasic_set_data_var_array(akt_data_array_var, akt_data_array_pos, value);
          if(failed==0) akt_data_array_pos++;
        }
      }
      else ok=1;
    }while(ok==0);
  }

  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
//------------------------------------------------------------
// Statement : "DATAW"
// Syntax    : DATAW<int_expr>[,<int_expr>...]
// Return    : <void>
//------------------------------------------------------------
static void dataw_statement(void)
{
  int value=0, ok=0;
  int hi,lo;

  accept(TOKENIZER_DATAW);
  value=expr();
  if((value<0) || (value>0xFFFF)) failed=UBASIC_INVALID_DATA;
  if(akt_data_array_var<0) failed=UBASIC_OVERFLOW;

  if(failed==0) {
    hi=((value>>8)&0xFF);
    lo=(value&0xFF);
    ubasic_set_data_var_array(akt_data_array_var, akt_data_array_pos, hi);
    if(failed==0) {
      akt_data_array_pos++;
      ubasic_set_data_var_array(akt_data_array_var, akt_data_array_pos, lo);
      if(failed==0) akt_data_array_pos++;
    }

    do {
      if(tokenizer_token()==TOKENIZER_COMMA) {
        value=0;
        accept(TOKENIZER_COMMA);
        value=expr();
        if((value<0) || (value>0xFFFF)) failed=UBASIC_INVALID_DATA;
        if(failed==0) {
          hi=((value>>8)&0xFF);
          lo=(value&0xFF);
          ubasic_set_data_var_array(akt_data_array_var, akt_data_array_pos, hi);
          if(failed==0) {
            akt_data_array_pos++;
            ubasic_set_data_var_array(akt_data_array_var, akt_data_array_pos, lo);
            if(failed==0) akt_data_array_pos++;
          }
        }
      }
      else ok=1;
    }while(ok==0);
  }

  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}



//------------------------------------------------------------
// NEW <INTEGER> STATEMENTS BY UB
//------------------------------------------------------------

//------------------------------------------------------------
// Statement : "ABS"
// Syntax    : ABS(<int_expr>)
// Return    : <int>
//------------------------------------------------------------
static int cmd_abs(void)
{
  int ret_value=0;

  accept(TOKENIZER_ABS);
  accept(TOKENIZER_LEFTPAREN);
  ret_value=expr();
  accept(TOKENIZER_RIGHTPAREN);
  if(ret_value<0) ret_value=0-ret_value;

  return ret_value;
}





//------------------------------------------------------------
// NEW <STRING> STATEMENTS BY UB
//------------------------------------------------------------


//------------------------------------------------------------
// Statement : "RIGHTSTR"
// Syntax    : RIGHT$(<str_expr>,<int_expr>)
// Return    : <str>
//------------------------------------------------------------
static char *cmd_rightstr(void)
{
  char *ret_wert,*s;
  int count,length,n,start;

  strbuf2[0]=0x00;
  ret_wert=(char*)&strbuf2[0];

  accept(TOKENIZER_RIGHTSTR);
  accept(TOKENIZER_LEFTPAREN);
  s=strexpr();
  length=strlen(s);
  accept(TOKENIZER_COMMA);
  count=expr();
  if(count<=0) failed=UBASIC_INVALID_DATA;
  accept(TOKENIZER_RIGHTPAREN);
  if((failed==0) && (length>0)) {
    if(count>=length) {
      ret_wert=s;
    }
    else {
      start=length-count;
      for(n=0;n<count;n++) {
        strbuf2[n]=s[n+start];
      }
      strbuf2[n]=0x00;
      ret_wert=(char*)&strbuf2[0];
    }
  }

  return ret_wert;
}




//------------------------------------------------------------
// END OF NEW STATEMENTS BY UB
//------------------------------------------------------------




//------------------------------------------------------------
// search target statement
//------------------------------------------------------------
static void statement(void)
{
  int token;

  if(failed!=0) return;

  token = tokenizer_token();

  switch(token) {
  case TOKENIZER_PRINT:
    print_statement();
    break;
  case TOKENIZER_IF:
    if_statement();
    break;
  case TOKENIZER_GOTO:
    goto_statement();
    break;
  case TOKENIZER_GOSUB:
    gosub_statement();
    break;
  case TOKENIZER_RETURN:
    return_statement();
    break;
  case TOKENIZER_FOR:
    for_statement();
    break;
  case TOKENIZER_NEXT:
    next_statement();
    break;
  case TOKENIZER_END:
    end_statement();
    break;
  case TOKENIZER_LET:
    accept(TOKENIZER_LET);
    /* Fall through. */
  case TOKENIZER_VARIABLE:
  case TOKENIZER_STRINGVAR:
  case TOKENIZER_VARARRAY:
  case TOKENIZER_STRINGARRAY:
  case TOKENIZER_DATAARRAY:
    let_statement();
    break;
  //----------------------------------------
  // NEW TOKENS WIHTOUT RETURN VALUES BY UB
  //----------------------------------------
  case TOKENIZER_REM :
    rem_statement();
    break;
  case TOKENIZER_DIM :
    dim_statement();
    break;
  case TOKENIZER_DATA :
    data_statement();
    break;
  case TOKENIZER_DATAW :
    dataw_statement();
    break;
  //----------------------------------------
  default:
    failed=UBASIC_INVALID_KEYWORD;
  }
}
//------------------------------------------------------------
static void linenum_statement(void)
{
  if(failed==0) {
    if(tokenizer_num()!=0) akt_linenum = tokenizer_num();
  }
  accept(TOKENIZER_LINENUM);
  statement();
  return;
}
//------------------------------------------------------------
void ubasic_run(void)
{
  if(tokenizer_finished()) {
    return;
  }

  linenum_statement();
}
//------------------------------------------------------------
int ubasic_finished(void)
{
  return ended || tokenizer_finished() || failed;
}
//------------------------------------------------------------
void ubasic_set_int_var(int varnum, int value)
{
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    int_var[varnum] = value;
  }
  else failed=UBASIC_INVALID_DATA;
}
//------------------------------------------------------------
int ubasic_get_int_var(int varnum)
{
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    return int_var[varnum];
  }
  else failed=UBASIC_INVALID_DATA;

  return 0;
}
//------------------------------------------------------------



//------------------------------------------------------------
// NEW FUNCTIONS BY UB
//------------------------------------------------------------


//------------------------------------------------------------
void ubasic_set_str_var(int varnum, char *ptr)
{
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    strcpy(str_var[varnum], ptr);
  }
  else failed=UBASIC_INVALID_DATA;
}
//------------------------------------------------------------
char *ubasic_get_str_var(int varnum)
{
  string[0]=0x00;
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    strcpy(string,str_var[varnum]);
  }
  else failed=UBASIC_INVALID_DATA;

  return string;
}
//------------------------------------------------------------
void ubasic_set_int_var_array(int varnum, int varpos, int value)
{
  int adr;

  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((varpos>=0) && (varpos<int_var_array[varnum].lenght)) {
      adr=varpos+int_var_array[varnum].start_adr;
      int_var_array_data[adr]=value;
    }
    else failed=UBASIC_OVERFLOW;
    if(varpos<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;
}
//------------------------------------------------------------
int ubasic_get_int_var_array(int varnum, int varpos)
{
  int adr;

  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((varpos>=0) && (varpos<int_var_array[varnum].lenght)) {
      adr=varpos+int_var_array[varnum].start_adr;
      return int_var_array_data[adr];
    }
    else failed=UBASIC_OVERFLOW;
    if(varpos<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;

  return 0;
}
//------------------------------------------------------------
void ubasic_set_str_var_array(int varnum, int varpos, char *ptr)
{
  int adr;

  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((varpos>=0) && (varpos<str_var_array[varnum].lenght)) {
      adr=varpos+str_var_array[varnum].start_adr;
      strcpy(str_var_array_data[adr], ptr);
    }
    else failed=UBASIC_OVERFLOW;
    if(varpos<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;
}
//------------------------------------------------------------
char *ubasic_get_str_var_array(int varnum, int varpos)
{
  int adr;

  string[0]=0x00;
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((varpos>=0) && (varpos<str_var_array[varnum].lenght)) {
      adr=varpos+str_var_array[varnum].start_adr;
      strcpy(string, str_var_array_data[adr]);
    }
    else failed=UBASIC_OVERFLOW;
    if(varpos<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;

  return string;
}
//------------------------------------------------------------
void ubasic_set_data_var_array(int varnum, int varpos, int value)
{
  int adr;

  if((value<0) || (value>255)) {
    failed=UBASIC_INVALID_DATA;
    return;
  }

  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((varpos>=0) && (varpos<data_var_array[varnum].lenght)) {
      adr=varpos+data_var_array[varnum].start_adr;
      data_var_array_data[adr]=value;
    }
    else failed=UBASIC_OVERFLOW;
    if(varpos<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;
}
//------------------------------------------------------------
int ubasic_get_data_var_array(int varnum, int varpos)
{
  int adr;

  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((varpos>=0) && (varpos<data_var_array[varnum].lenght)) {
      adr=varpos+data_var_array[varnum].start_adr;
      return data_var_array_data[adr];
    }
    else failed=UBASIC_OVERFLOW;
    if(varpos<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;

  return 0;
}
//------------------------------------------------------------
int ubasic_check_data_var_array(int varnum, int varpos, int length)
{
  int adr=0;

  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((varpos>=0) && (varpos+length<=data_var_array[varnum].lenght)) {
      adr=varpos+data_var_array[varnum].start_adr;
      return adr;
    }
    else failed=UBASIC_OVERFLOW;
    if(varpos<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;

  return adr;
}
//------------------------------------------------------------
void ubasic_set_datas_var_array(int varnum, int varpos, char *ptr)
{
  int adr,n;
  char wert,*str_ptr;

  str_ptr=ptr;
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((varpos>=0) && (varpos<data_var_array[varnum].lenght)) {
      adr=varpos+data_var_array[varnum].start_adr;
      for(n=0;n<MAX_STRINGLEN;n++) {
    	if(n+varpos+1>=data_var_array[varnum].lenght) {
    	  failed=UBASIC_OVERFLOW;
    	  break;
    	}
    	else {
    	  wert=*str_ptr;
    	  data_var_array_data[adr]=wert;
    	  if(wert==0x00) break;
    	  adr++;
    	  str_ptr++;
    	}
    	data_var_array_data[adr]=0x00;
      }
    }
    else failed=UBASIC_OVERFLOW;
    if(varpos<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;
}
//------------------------------------------------------------
char *ubasic_get_datas_var_array(int varnum, int varpos)
{
  int adr,n;
  char wert;

  string[0]=0x00;
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((varpos>=0) && (varpos<data_var_array[varnum].lenght)) {
      adr=varpos+data_var_array[varnum].start_adr;
      for(n=0;n<MAX_STRINGLEN;n++) {
      	if(n+varpos+1>=data_var_array[varnum].lenght) {
      	  failed=UBASIC_OVERFLOW;
      	  break;
      	}
      	else {
      	  wert=data_var_array_data[adr];
      	  string[n]=wert;
      	  if(wert==0x00) break;
      	  adr++;
      	}
      }
      string[n]=0x00;
    }
    else failed=UBASIC_OVERFLOW;
    if(varpos<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;

  return string;
}
//------------------------------------------------------------
void ubasic_dim_int_array(int varnum, int length)
{
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((length>0) && (int_var_array[varnum].lenght==0)) {
      if(int_var_array_ptr+length<MAX_INT_ARRAY_SIZE) {
        int_var_array[varnum].start_adr=int_var_array_ptr;
        int_var_array[varnum].lenght=length;
        int_var_array_ptr+=length;
      }
      else failed=UBASIC_OUT_OF_MEMORY;
    }
    else failed=UBASIC_OVERFLOW;
    if(length<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;
}
//------------------------------------------------------------
void ubasic_dim_str_array(int varnum, int length)
{
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((length>0) && (str_var_array[varnum].lenght==0)) {
      if(str_var_array_ptr+length<MAX_STR_ARRAY_SIZE) {
        str_var_array[varnum].start_adr=str_var_array_ptr;
        str_var_array[varnum].lenght=length;
        str_var_array_ptr+=length;
      }
      else failed=UBASIC_OUT_OF_MEMORY;
    }
    else failed=UBASIC_OVERFLOW;
    if(length<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;
}
//------------------------------------------------------------
void ubasic_dim_data_array(int varnum, int length)
{
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    if((length>0) && (data_var_array[varnum].lenght==0)) {
      if(data_var_array_ptr+length<MAX_DATA_ARRAY_SIZE) {
        data_var_array[varnum].start_adr=data_var_array_ptr;
        data_var_array[varnum].lenght=length;
        data_var_array_ptr+=length;
        akt_data_array_var=varnum;
        akt_data_array_pos=0;
      }
      else failed=UBASIC_OUT_OF_MEMORY;
    }
    else failed=UBASIC_OVERFLOW;
    if(length<0) failed=UBASIC_INVALID_DATA;
  }
  else failed=UBASIC_INVALID_DATA;
}
//------------------------------------------------------------
static void skipelse(void)
{
  do {
    tokenizer_next();
  } while(tokenizer_token() != TOKENIZER_CR &&
          tokenizer_token() != TOKENIZER_ENDOFINPUT);

  if(tokenizer_token() == TOKENIZER_CR) {
    accept(TOKENIZER_CR);
  }
}
//------------------------------------------------------------
int ubasic_failed(void)
{
  return failed;
}
//------------------------------------------------------------
int ubasic_linenum(void)
{
  return akt_linenum;
}
//------------------------------------------------------------
int ubasic_getkeyword(char token, char *ptr)
{
  int ret_wert;

  ret_wert=tokenizer_getkeyword(token, ptr);

  return ret_wert;
}
//------------------------------------------------------------
int ubasic_getramstatus(char mode)
{
  if(mode==0) return MAX_INT_ARRAY_SIZE;
  if(mode==1) return int_var_array_ptr;
  if(mode==2) return MAX_STR_ARRAY_SIZE;
  if(mode==3) return str_var_array_ptr;
  if(mode==4) return MAX_DATA_ARRAY_SIZE;
  if(mode==5) return data_var_array_ptr;

  return 0;
}
